#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "life.h"
#include "lookup.h"
#include "util.h"

/* Private Function Prototypes */

cage_type*  birth_cage(int32 x, int32 y);
void        get_block_text(uint16 block, char buf[8][9], int32 xoffset, int32 yoffset);
cage_type*  find_cage(int32 x, int32 y);
cage_type*  find_rattle_birth(int32 x, int32 y);
void        kill_cage(cage_type* cage);
void        move_cage(cage_type* cage, cage_type** old_list, cage_type** new_list);
void        predict_cage(cage_type* cage);
void        rattle_cage(cage_type* cage, boolean cage_changed);
void        rattle_neighbors(cage_type* cage);
void        tranquilize_cage(cage_type* cage);
void        update_onscreen_flag(cage_type* cage);

/* Defined in the GUI code */
void draw_life_block(uint16 block, int32 xstart, int32 ystart, boolean full_update);

/* Constants */

static uint8 bitcount[0x100] = {
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

/* Public Globals */

uint32      tick;
boolean     parity;            /* on an odd (1) or even (0) tick? */
uint32      population;

char        pattern_description[DESCRIPTION_MAX_LINES][DESCRIPTION_MAX_COLS+1];
int32       desc_num_lines;

/* Private Globals */

/* doubly-linked lists */
cage_type*  active_cages;
cage_type*  sleeping_cages;
cage_type*  predictable_cages_onscreen;
cage_type*  predictable_cages_offscreen;

static cage_type*  dead_cages;               /* singly-linked list */
static cage_type*  cage_hash[XY_HASH_SIZE];  /* for finding new neighbors */
static uint32      sleeping_population, predictable_population[2];

/* The frontend tells us about the viewport dimensions, so that we know which cells should
 * be drawn onscreen. */
static int32       viewport_xstart = -1, viewport_ystart = -1;
static int32       viewport_xend   = -1, viewport_yend   = -1;

/* Public Functions */

/* Completely clear the universe, emptying all cage lists, emptying the hash, resetting ticks to 0
 * and pattern_description to NULL.
 */
void clear_world(void)
{
    /* Clear all cage lists */
    clear_cage_list(&active_cages);
    clear_cage_list(&sleeping_cages);
    clear_cage_list(&predictable_cages_onscreen);
    clear_cage_list(&predictable_cages_offscreen);
    clear_cage_list(&dead_cages);

    /* Zero-out the hash */
    memset(cage_hash, 0, XY_HASH_SIZE * sizeof(cage_type*));

    /* Set ticks and population to 0 */
    tick = parity = population = sleeping_population = 0;
    predictable_population[0] = predictable_population[1] = 0;

    /* Clear the description */
    desc_num_lines = 0;
}

/* Deallocate the given linked list of cages and set to NULL.
 */
void clear_cage_list(cage_type** start)
{
    cage_type*  p;
    cage_type*  next;

    for (p=*start; p; p=next) {
        next = p->next;
        free(p);
    }
    *start = NULL;
}

/* Set, unset, or toggle the cell at the given XY coordinates.
 */
void draw_cell(int32 x, int32 y, draw_method_type draw_method)
{
    cage_type*  cage;
    int32       cagex, cagey;
    int32       xoffset, yoffset;
    uint16*     block;
    uint16      bitmask;

    /* Check if coordinates are out of bounds */
    if (x < 0 || x >= WORLD_SIZE || y < 0 || y >= WORLD_SIZE)
        return;

    /* Find the cage containing this cell, rattle it, birth it if necessary */
    cagex = x / CAGE_SIZE;
    cagey = y / CAGE_SIZE;
    cage = find_cage(cagex, cagey);
    if (cage)
        rattle_cage(cage, TRUE);
    else {
        if (draw_method == DRAW_UNSET)
            return;
        else
            cage = birth_cage(cagex, cagey);
    }
    rattle_neighbors(cage);

    /* Extract the uint16 representing the appropriate 4x4 block; get offsets relative
     * to the block.
     */
    xoffset = x % CAGE_SIZE;
    yoffset = y % CAGE_SIZE;
    if (yoffset < 4) {
        if (xoffset < 4)
            block = &(cage->bnw[parity]);
        else
            block = &(cage->bne[parity]);
    } else {
        if (xoffset < 4)
            block = &(cage->bsw[parity]);
        else
            block = &(cage->bse[parity]);
    }
    yoffset %= 4;
    xoffset %= 4;

    /* Set or unset the appropriate bit */
    bitmask = 1 << (yoffset*4 + xoffset);
    if (draw_method == DRAW_SET) {
        if (!(*block & bitmask)) {
            *block |= bitmask;
            population++;
        }
    }
    else {
        if (*block & bitmask) {
            *block &= ~bitmask;
            population--;
        }
    }
}

/* Load a pattern from the given file, returning true if successful.
 */
load_result_type load_pattern(const char* path)
{
    FILE*   f;
    char*   line;
    char*   desc;
    char*   p;
    double  version;
    int32   x, y, curx;

    /* Open the file */
    f = fopen(path, "r");
    if (!f)
        return ((errno == ENOENT) ? LOAD_NOT_FOUND : LOAD_SYS_ERROR);

    /* Make sure this is a proper .lif file. Check pattern-file version, make sure its at
     * or below PATTERNS_MAX_VERSION
     */
    line = dgets(f);
    if (!line || !STR_STARTS_WITH(line, "#Life"))
        return LOAD_INVALID_FORMAT;
    if (sscanf(line+5, "%lf", &version) != 1 || version > PATTERNS_MAX_VERSION)
        return LOAD_INVALID_FORMAT;
    free(line);

    /* The file appears to be legit, so clear the world */
    clear_world();

    /* Load the file */
    line = dgets(f);
    while (line) {
        if (STR_STARTS_WITH(line, "#D")) {
            /* Load a description line, unless we are at max lines */
            if (desc_num_lines < DESCRIPTION_MAX_LINES) {
                desc = line+2;
                trim_whitespace(desc);
                if (strlen(desc) > DESCRIPTION_MAX_COLS)
                    desc[DESCRIPTION_MAX_COLS] = '\0';
                strcpy(pattern_description[desc_num_lines++], desc);
            }
        } else if (STR_STARTS_WITH(line, "#P")) {
            /* Load a block of cell data */
            if (sscanf(line+2, "%d%d", &x, &y) != 2)
                x = y = 0;
            x += WORLD_SIZE / 2;
            y += WORLD_SIZE / 2;
            free(line);
            while ((line = dgets(f)) != NULL && !STR_STARTS_WITH(line, "#")) {
                /* Load this line of data, unless we are out of bounds or the line is empty */
                if (y < WORLD_SIZE && line[0] != '\0') {
                    for (p=line, curx=x; *p; p++, curx++) {
                        if (curx < WORLD_SIZE && *p == '*')
                            draw_cell(curx, y, DRAW_SET);
                    }
                    y++;
                }
                free(line);
            }
            continue;
        }

        free(line);
        line = dgets(f);
    }

    fclose(f);
    return LOAD_SUCCESS;
}

/* Loops through all cages in the Life world, returning NULL at the end. This function should only
 * be used by one caller at a time, and should be used until NULL. It is safe to move cages between
 * linked lists while this loop is active.
 */
cage_type* loop_cages(void)
{
    static cage_type*  pos = NULL;
    static int32       hash_pos = -1;

    if (pos)
        pos = pos->hash_next;

    if (!pos) {
        hash_pos++;
        while (hash_pos < XY_HASH_SIZE && !cage_hash[hash_pos])
            hash_pos++;
        if (hash_pos == XY_HASH_SIZE) {
            hash_pos = -1;
            return NULL;
        }
        pos = cage_hash[hash_pos];
    }

    return pos;
}

/* Loops through all onscreen (contained in the viewport specified by set_viewport) cages in the
 * Life world, returning NULL at the end. This function should only be used by one caller at a
 * time, and should be used until NULL. It is assumed that cages will not be moved between linked
 * lists while this loop is active.
 */
cage_type* loop_cages_onscreen(void)
{
    static cage_type*  pos = NULL;
    static boolean     processed_sleepers, processed_preds;

    if (pos)
        pos = pos->next;
    else {
        pos = active_cages;
        processed_sleepers = processed_preds = FALSE;
    }
    if (!pos && !processed_sleepers) {
        processed_sleepers = TRUE;
        pos = sleeping_cages;
    }
    if (!pos && !processed_preds) {
        processed_preds = TRUE;
        pos = predictable_cages_onscreen;
    }

    while (pos && !IS_ONSCREEN(pos)) {
        pos = pos->next;
        if (!pos && !processed_sleepers) {
            processed_sleepers = TRUE;
            pos = sleeping_cages;
        }
        if (!pos && !processed_preds) {
            processed_preds = TRUE;
            pos = predictable_cages_onscreen;
        }
    }

    return pos;
}

/* Apply the given bitmasks to the four corners of the cage to reduce its contents, zeroing any
 * cells that are not 1 in the mask.
 */
void mask_cage(cage_type* cage, uint16 nw_mask, uint16 ne_mask, uint16 sw_mask, uint16 se_mask)
{
    uint16  nw, ne, sw, se;
    uint32  old_pop, new_pop;

    nw = cage->bnw[parity];
    ne = cage->bne[parity];
    sw = cage->bsw[parity];
    se = cage->bse[parity];
    old_pop = POPULATION_COUNT(nw, ne, sw, se);
    nw &= nw_mask;
    ne &= ne_mask;
    sw &= sw_mask;
    se &= se_mask;
    new_pop = POPULATION_COUNT(nw, ne, sw, se);

    if (old_pop != new_pop) {
        cage->bnw[parity] = nw;
        cage->bne[parity] = ne;
        cage->bsw[parity] = sw;
        cage->bse[parity] = se;
        population -= old_pop - new_pop;
        rattle_cage(cage, TRUE);
        rattle_neighbors(cage);
    }
}

/* Calculate the next Life generation, and update the pixmap via calls to the external function
 * draw_life_block.
 */
void next_tick(void)
{
    cage_type*  cage;
    cage_type*  next = NULL;
    boolean     oscillating_to_stable;
    int32       turns_stable, turns_oscillating;
    int32       t1, t2;
    int32       x, y;
    uint16      nw, ne, sw, se;
    uint16      old_nw, old_ne, old_sw, old_se;
    uint16      new_nw, new_ne, new_sw, new_se;
    uint16      neigh_nnw, neigh_nne, neigh_ssw, neigh_sse, neigh_wnw, neigh_wsw,
                neigh_ene, neigh_ese, neigh_nw, neigh_ne, neigh_sw, neigh_se;

    t1 = parity;
    t2 = !parity;
    population = sleeping_population + predictable_population[t2];

    for (cage=active_cages; cage; cage=next) {
        /* Store next-pointer in case we end up moving this cage to another linked list */
        next = cage->next;

        /* Extract the four 4x4 blocks in this cage, past and present, as well as any neighboring
         * 4x4 blocks that we'll need to know about.
         */
        nw = cage->bnw[t1];
        ne = cage->bne[t1];
        sw = cage->bsw[t1];
        se = cage->bse[t1];
        old_nw = cage->bnw[t2];
        old_ne = cage->bne[t2];
        old_sw = cage->bsw[t2];
        old_se = cage->bse[t2];
        neigh_nnw = neigh_nne = neigh_ssw = neigh_sse = neigh_wnw = neigh_wsw =
            neigh_ene = neigh_ese = neigh_nw = neigh_ne = neigh_sw = neigh_se = 0;
        if (cage->north) {
            neigh_nnw = cage->north->bsw[t1];
            neigh_nne = cage->north->bse[t1];
        }
        if (cage->south) {
            neigh_ssw = cage->south->bnw[t1];
            neigh_sse = cage->south->bne[t1];
        }
        if (cage->west) {
            neigh_wnw = cage->west->bne[t1];
            neigh_wsw = cage->west->bse[t1];
        }
        if (cage->east) {
            neigh_ene = cage->east->bnw[t1];
            neigh_ese = cage->east->bsw[t1];
        }
        if (cage->nw)
            neigh_nw = cage->nw->bse[t1];
        if (cage->ne)
            neigh_ne = cage->ne->bsw[t1];
        if (cage->sw)
            neigh_sw = cage->sw->bne[t1];
        if (cage->se)
            neigh_se = cage->se->bnw[t1];

        /* Calculate the new 4x4 blocks for this cage */
        new_nw =
            ul_lookup[(nw & 0x0777) | (neigh_nnw & 0x7000) | (neigh_wnw & 0x0888) | (neigh_nw & 0x8000)]   |
            ur_lookup[(nw & 0x0EEE) | (neigh_nnw & 0xE000) | (ne & 0x0111)        | (neigh_nne & 0x1000)]  |
            ll_lookup[(nw & 0x7770) | (sw & 0x0007)        | (neigh_wnw & 0x8880) | (neigh_wsw & 0x0008)]  |
            lr_lookup[(nw & 0xEEE0) | (sw & 0x000E)        | (ne & 0x1110)        | (se & 0x0001)];
        new_ne =
            ul_lookup[(ne & 0x0777) | (neigh_nne & 0x7000) | (nw & 0x0888)        | (neigh_nnw & 0x8000)]  |
            ur_lookup[(ne & 0x0EEE) | (neigh_nne & 0xE000) | (neigh_ene & 0x0111) | (neigh_ne & 0x1000)]   |
            ll_lookup[(ne & 0x7770) | (se & 0x0007)        | (nw & 0x8880)        | (sw & 0x0008)]         |
            lr_lookup[(ne & 0xEEE0) | (se & 0x000E)        | (neigh_ene & 0x1110) | (neigh_ese & 0x0001)];
        new_sw =
            ul_lookup[(sw & 0x0777) | (nw & 0x7000)        | (neigh_wsw & 0x0888) | (neigh_wnw & 0x8000)]  |
            ur_lookup[(sw & 0x0EEE) | (nw & 0xE000)        | (se & 0x0111)        | (ne & 0x1000)]         |
            ll_lookup[(sw & 0x7770) | (neigh_ssw & 0x0007) | (neigh_wsw & 0x8880) | (neigh_sw & 0x0008)]   |
            lr_lookup[(sw & 0xEEE0) | (neigh_ssw & 0x000E) | (se & 0x1110)        | (neigh_sse & 0x0001)];
        new_se =
            ul_lookup[(se & 0x0777) | (ne & 0x7000)        | (sw & 0x0888)        | (nw & 0x8000)]         |
            ur_lookup[(se & 0x0EEE) | (ne & 0xE000)        | (neigh_ese & 0x0111) | (neigh_ene & 0x1000)]  |
            ll_lookup[(se & 0x7770) | (neigh_sse & 0x0007) | (sw & 0x8880)        | (neigh_ssw & 0x0008)]  |
            lr_lookup[(se & 0xEEE0) | (neigh_sse & 0x000E) | (neigh_ese & 0x1110) | (neigh_se & 0x0001)];

        /* Update population count */
        population += POPULATION_COUNT(new_nw, new_ne, new_sw, new_se);

        /* Check if the cage is stable or oscillating, and update TURNS_STABLE and
         * TURNS_OSCILLATING accordingly. If it's been stable awhile, move it to sleeping_cages or
         * dead_cages (if empty). If it's been oscillating awhile, move it to predictable_cages.
         */
        turns_oscillating = TURNS_OSCILLATING(cage);
        oscillating_to_stable = FALSE;
        if (new_nw == nw && new_ne == ne && new_sw == sw && new_se == se) {
            turns_stable = TURNS_STABLE(cage);
            if (turns_stable == MAX_TURNS_STABLE) {
                if (!new_nw && !new_ne && !new_sw && !new_se)
                    kill_cage(cage);
                else
                    tranquilize_cage(cage);
                continue;           /* stable for quite awhile now: no need to redraw, apply new blocks, */
                                    /* or rattle neighbors */
            } else {
                INC_TURNS_STABLE(cage);
                if (turns_stable)   /* stable for two ticks now: no need to redraw, apply new blocks, */
                    continue;       /* or rattle neighbors */
                turns_stable++;
                if (turns_oscillating) {
                    RESET_TURNS_OSCILLATING(cage);
                    turns_oscillating = 0;
                    oscillating_to_stable = TRUE;
                }
            }
        } else {
            RESET_TURNS_STABLE(cage);
            turns_stable = 0;
            if (new_nw == old_nw && new_ne == old_ne && new_sw == old_sw && new_se == old_se) {
                if (turns_oscillating == MAX_TURNS_OSCILLATING) {
                    predict_cage(cage);
                    continue;   /* no need to redraw (see below loop), apply new blocks, or rattle neighbors */
                } else {
                    INC_TURNS_OSCILLATING(cage);
                    turns_oscillating++;
                }
            }
            else {
                RESET_TURNS_OSCILLATING(cage);
                turns_oscillating = 0;
            }
        }

        /* Update the Life pixmap if necessary */
        if (!turns_stable && IS_ONSCREEN(cage)) {
            x = cage->x * CAGE_SIZE;
            y = cage->y * CAGE_SIZE;
            if (new_nw != nw)
                draw_life_block(new_nw, x, y, FALSE);
            if (new_ne != ne)
                draw_life_block(new_ne, x + BLOCK_SIZE, y, FALSE);
            if (new_sw != sw)
                draw_life_block(new_sw, x, y + BLOCK_SIZE, FALSE);
            if (new_se != se)
                draw_life_block(new_se, x + BLOCK_SIZE, y + BLOCK_SIZE, FALSE);
        }

        if (turns_oscillating)
            continue;   /* no need to apply blocks or rattle neighbors */

        /* Apply the new blocks */
        cage->bnw[t2] = new_nw;
        cage->bne[t2] = new_ne;
        cage->bsw[t2] = new_sw;
        cage->bse[t2] = new_se;

        /* No need to rattle neighbors if we've been stable for one tick, unless we were
         * previously oscillating */
        if (turns_stable && !oscillating_to_stable)
            continue;

        /* Rattle/find/create neighbors as needed */
        x = cage->x;
        y = cage->y;
        if ((new_nw & NORTH_EDGE_MASK) != (nw & NORTH_EDGE_MASK) ||
            (new_ne & NORTH_EDGE_MASK) != (ne & NORTH_EDGE_MASK) ||
            (new_nw & NORTH_EDGE_MASK) != (old_nw & NORTH_EDGE_MASK) ||
            (new_ne & NORTH_EDGE_MASK) != (old_ne & NORTH_EDGE_MASK)) {
            if (cage->north)
                rattle_cage(cage->north, FALSE);
            else {
                cage->north = find_rattle_birth(x, y-1);
                if (cage->north)
                    cage->north->south = cage;
            }
        }
        if ((new_sw & SOUTH_EDGE_MASK) != (sw & SOUTH_EDGE_MASK) ||
            (new_se & SOUTH_EDGE_MASK) != (se & SOUTH_EDGE_MASK) ||
            (new_sw & SOUTH_EDGE_MASK) != (old_sw & SOUTH_EDGE_MASK) ||
            (new_se & SOUTH_EDGE_MASK) != (old_se & SOUTH_EDGE_MASK)) {
            if (cage->south)
                rattle_cage(cage->south, FALSE);
            else {
                cage->south = find_rattle_birth(x, y+1);
                if (cage->south)
                    cage->south->north = cage;
            }
        }
        if ((new_nw & WEST_EDGE_MASK) != (nw & WEST_EDGE_MASK) ||
            (new_sw & WEST_EDGE_MASK) != (sw & WEST_EDGE_MASK) ||
            (new_nw & WEST_EDGE_MASK) != (old_nw & WEST_EDGE_MASK) ||
            (new_sw & WEST_EDGE_MASK) != (old_sw & WEST_EDGE_MASK)) {
            if (cage->west)
                rattle_cage(cage->west, FALSE);
            else {
                cage->west = find_rattle_birth(x-1, y);
                if (cage->west)
                    cage->west->east = cage;
            }
        }
        if ((new_ne & EAST_EDGE_MASK) != (ne & EAST_EDGE_MASK) ||
            (new_se & EAST_EDGE_MASK) != (se & EAST_EDGE_MASK) ||
            (new_ne & EAST_EDGE_MASK) != (old_ne & EAST_EDGE_MASK) ||
            (new_se & EAST_EDGE_MASK) != (old_se & EAST_EDGE_MASK)) {
            if (cage->east)
                rattle_cage(cage->east, FALSE);
            else {
                cage->east = find_rattle_birth(x+1, y);
                if (cage->east)
                    cage->east->west = cage;
            }
        }
        if ((new_nw & NW_CORNER_MASK) != (nw & NW_CORNER_MASK) ||
            (new_nw & NW_CORNER_MASK) != (old_nw & NW_CORNER_MASK)) {
            if (cage->nw)
                rattle_cage(cage->nw, FALSE);
            else {
                cage->nw = find_rattle_birth(x-1, y-1);
                if (cage->nw)
                    cage->nw->se = cage;
            }
        }
        if ((new_ne & NE_CORNER_MASK) != (ne & NE_CORNER_MASK) ||
            (new_ne & NE_CORNER_MASK) != (old_ne & NE_CORNER_MASK)) {
            if (cage->ne)
                rattle_cage(cage->ne, FALSE);
            else {
                cage->ne = find_rattle_birth(x+1, y-1);
                if (cage->ne)
                    cage->ne->sw = cage;
            }
        }
        if ((new_sw & SW_CORNER_MASK) != (sw & SW_CORNER_MASK) ||
            (new_sw & SW_CORNER_MASK) != (old_sw & SW_CORNER_MASK)) {
            if (cage->sw)
                rattle_cage(cage->sw, FALSE);
            else {
                cage->sw = find_rattle_birth(x-1, y+1);
                if (cage->sw)
                    cage->sw->ne = cage;
            }
        }
        if ((new_se & SE_CORNER_MASK) != (se & SE_CORNER_MASK) ||
            (new_se & SE_CORNER_MASK) != (old_se & SE_CORNER_MASK)) {
            if (cage->se)
                rattle_cage(cage->se, FALSE);
            else {
                cage->se = find_rattle_birth(x+1, y+1);
                if (cage->se)
                    cage->se->nw = cage;
            }
        }
    }

    /* Update Life pixmap for on-screen predictable cages */
    for (cage=predictable_cages_onscreen; cage; cage=cage->next) {
        x = cage->x * CAGE_SIZE;
        y = cage->y * CAGE_SIZE;
        draw_life_block(cage->bnw[t2], x, y, FALSE);
        draw_life_block(cage->bne[t2], x + BLOCK_SIZE, y, FALSE);
        draw_life_block(cage->bsw[t2], x, y + BLOCK_SIZE, FALSE);
        draw_life_block(cage->bse[t2], x + BLOCK_SIZE, y + BLOCK_SIZE, FALSE);
    }

    tick++;
    parity = !parity;
}

/* Save the current pattern to the given file, returning true if successful, false if a system
 * error occurred in opening or writing the file.
 */
boolean save_pattern(const char* path)
{
    FILE*       f;
    cage_type*  c;
    char        chunk[8][9];
    int32       startx, starty;
    int32       first_row, last_row, first_col;
    int32       x, y;
    int32       i;

    /* Open the file for writing and print the standard header */
    f = fopen(path, "w");
    if (!f)
        return FALSE;
    fprintf(f, "#Life %.2f\r\n", PATTERNS_MAX_VERSION);

    /* Output the description, if any */
    for (i=0; i < desc_num_lines; i++)
        fprintf(f, "#D %s\r\n", pattern_description[i]);

    /* Normal rules */
    fprintf(f, "#N\r\n");

    /* Output the cage data */
    for (y=0; y < 8; y++)
        chunk[y][8] = '\0';
    while ((c = loop_cages()) != NULL) {
        if (IS_EMPTY(c))
            continue;

        /* Get the starting coords for the cage, relative to center of world */
        startx = (c->x * CAGE_SIZE) - WORLD_SIZE/2;
        starty = (c->y * CAGE_SIZE) - WORLD_SIZE/2;

        /* Fill the chunk with the data from the four 4x4 blocks */
        get_block_text(c->bnw[parity], chunk, 0, 0);
        get_block_text(c->bne[parity], chunk, 4, 0);
        get_block_text(c->bsw[parity], chunk, 0, 4);
        get_block_text(c->bse[parity], chunk, 4, 4);

        /* Trim off empty rows, empty columns, and trailing empty cells */
        for (y=0; y < 8 && STR_EQUAL(chunk[y], "........"); y++, starty++) {}
        first_row = y;
        for (y=7; y > first_row && STR_EQUAL(chunk[y], "........"); y--) {}
        last_row = y;
        for (x=0; x < 8; x++, startx++) {
            for (y=0; y < 8 && chunk[y][x] == '.'; y++) {}
            if (y < 8)
                break;
        }
        first_col = x;
        for (y=0; y < 8; y++) {
            for (x=7; x > first_col && chunk[y][x] == '.'; x--) { }
            chunk[y][x+1] = '\0';
        }

        /* Output chunk header and data */
        fprintf(f, "#P %d %d\r\n", startx, starty);
        for (y=first_row; y <= last_row; y++)
            fprintf(f, "%s\r\n", chunk[y] + first_col);
    }

    /* Close file, returning FALSE if there was a write error */
    if (fclose(f) != 0)
        return FALSE;

    return TRUE;
}

/* Set the current Life viewport (in cell coordinates), which is used to determine when to call
 * draw_life_block() in next_tick().
 */
void set_viewport(int32 xstart, int32 ystart, int32 xend, int32 yend)
{
    cage_type*  cage;
    cage_type*  next;
    cage_type*  org_pred_cages_offscreen;

    viewport_xstart = xstart;
    viewport_ystart = ystart;
    viewport_xend = xend;
    viewport_yend = yend;

    for (cage=active_cages; cage; cage=cage->next)
        update_onscreen_flag(cage);
    for (cage=sleeping_cages; cage; cage=cage->next)
        update_onscreen_flag(cage);
    org_pred_cages_offscreen = predictable_cages_offscreen;
    for (cage=predictable_cages_onscreen; cage; cage=next) {
        next = cage->next;
        update_onscreen_flag(cage);
        if (!IS_ONSCREEN(cage))
            move_cage(cage, &predictable_cages_onscreen, &predictable_cages_offscreen);
    }
    for (cage=org_pred_cages_offscreen; cage; cage=next) {
        next = cage->next;
        update_onscreen_flag(cage);
        if (IS_ONSCREEN(cage))
            move_cage(cage, &predictable_cages_offscreen, &predictable_cages_onscreen);
    }
}

/* Private Functions */

/* Allocate the cage at the given x, y position, find its neighbors, and insert it into
 * the active cages list and cage hash. Return a pointer to the new cage. If x and/or
 * y is out of bounds, return NULL.
 */
cage_type* birth_cage(int32 x, int32 y)
{
    cage_type*  cage;
    uint16      hash;

    /* Return NULL if x and/or y is out of bounds */
    if (x < 0 || x >= WORLD_CAGES || y < 0 || y >= WORLD_CAGES)
        return NULL;

    /* Allocate, set X and Y */
    if (dead_cages) {
        cage = dead_cages;
        dead_cages = dead_cages->next;
        memset(cage, 0, sizeof(cage_type));
    } else
        cage = safe_calloc(1, sizeof(cage_type));
    cage->x = x;
    cage->y = y;

    /* Set whether it's onscreen or offscreen */
    update_onscreen_flag(cage);

    /* Find allocated neighbors */
    cage->north = find_cage(x, y-1);
    if (cage->north)
        cage->north->south = cage;
    cage->south = find_cage(x, y+1);
    if (cage->south)
        cage->south->north = cage;
    cage->west = find_cage(x-1, y);
    if (cage->west)
        cage->west->east = cage;
    cage->east = find_cage(x+1, y);
    if (cage->east)
        cage->east->west = cage;
    cage->nw = find_cage(x-1, y-1);
    if (cage->nw)
        cage->nw->se = cage;
    cage->ne = find_cage(x+1, y-1);
    if (cage->ne)
        cage->ne->sw = cage;
    cage->sw = find_cage(x-1, y+1);
    if (cage->sw)
        cage->sw->ne = cage;
    cage->se = find_cage(x+1, y+1);
    if (cage->se)
        cage->se->nw = cage;

    /* Insert into active cages list */
    cage->prev = NULL;
    cage->next = active_cages;
    if (active_cages)
        active_cages->prev = cage;
    active_cages = cage;

    /* Insert into cage hash */
    hash = XY_HASH(x,y);
    cage->hash_prev = NULL;
    cage->hash_next = cage_hash[hash];
    if (cage_hash[hash])
        cage_hash[hash]->hash_prev = cage;
    cage_hash[hash] = cage;

    return cage;
}

/* Fill a quarter of the buffer buf with text representing the given 4x4 cell block, using the
 * given x/y offsets into the buffer.
 */
void get_block_text(uint16 block, char buf[8][9], int32 xoffset, int32 yoffset)
{
    int32  x, y;
    int32  bit;

    for (y=yoffset, bit=0; y < yoffset+4; y++) {
        for (x=xoffset; x < xoffset+4; x++, bit++)
            buf[y][x] = ((block & (1 << bit)) ? '*' : '.');
    }
}

/* Search the cage hash for the cage with the given XY coordinates. Return a pointer to
 * the cage, or NULL if it isn't allocated.
 */
cage_type* find_cage(int32 x, int32 y)
{
    uint16      hash;
    cage_type*  cage;

    if (x < 0 || x >= WORLD_CAGES || y < 0 || y >= WORLD_CAGES)
        return NULL;
    hash = XY_HASH(x,y);
    for (cage=cage_hash[hash];
         cage && !(cage->x == x && cage->y == y);
         cage = cage->hash_next) {}
    return cage;
}

/* Search the cage hash for the cage with the given XY coordinates. If found, rattle it
 * and return a pointer to it, otherwise birth and return a pointer to a new cage.
 */
cage_type* find_rattle_birth(int32 x, int32 y)
{
    cage_type* c;

    c = find_cage(x, y);
    if (c)
        rattle_cage(c, FALSE);
    else
        c = birth_cage(x, y);
    return c;
}

/* Move a cage into the dead_cages pool, deregister it from the hash, and set neighboring pointers
 * to it to NULL.
 */
void kill_cage(cage_type* cage)
{
    /* Set neighboring pointers to self to NULL */
    if (cage->north)
        cage->north->south = NULL;
    if (cage->south)
        cage->south->north = NULL;
    if (cage->west)
        cage->west->east = NULL;
    if (cage->east)
        cage->east->west = NULL;
    if (cage->nw)
        cage->nw->se = NULL;
    if (cage->ne)
        cage->ne->sw = NULL;
    if (cage->sw)
        cage->sw->ne = NULL;
    if (cage->se)
        cage->se->nw = NULL;

    /* Move into dead_cages list */
    if (cage->prev)
        cage->prev->next = cage->next;
    else
        active_cages = cage->next;
    if (cage->next)
        cage->next->prev = cage->prev;
    cage->next = dead_cages;
    dead_cages = cage;

    /* Deregister from the cage hash */
    if (cage->hash_prev)
        cage->hash_prev->hash_next = cage->hash_next;
    else
        cage_hash[XY_HASH(cage->x, cage->y)] = cage->hash_next;
    if (cage->hash_next)
        cage->hash_next->hash_prev = cage->hash_prev;
}

/* Move the given cage from its place in old_list to the front of new_list (both doubly-linked
 * lists).
 */
void move_cage(cage_type* cage, cage_type** old_list, cage_type** new_list)
{
    if (cage->prev)
        cage->prev->next = cage->next;
    else
        *old_list = cage->next;
    if (cage->next)
        cage->next->prev = cage->prev;
    cage->prev = NULL;
    cage->next = *new_list;
    if (*new_list)
        (*new_list)->prev = cage;
    *new_list = cage;
}

/* Flag a cage as predictable and move it to the appropriate list, adding its live cell counts for
 * odd and even ticks to predictable_population[2].
 */
void predict_cage(cage_type* cage)
{
    SET_PREDICTABLE(cage);
    move_cage(cage, &active_cages,
              (IS_ONSCREEN(cage) ? &predictable_cages_onscreen : &predictable_cages_offscreen));
    predictable_population[0] += POPULATION_COUNT(cage->bnw[0], cage->bne[0], cage->bsw[0],
                                                  cage->bse[0]);
    predictable_population[1] += POPULATION_COUNT(cage->bnw[1], cage->bne[1], cage->bsw[1],
                                                  cage->bse[1]);
}

/* If cage_changed is true, set TURNS_STABLE and TURNS_OSCILLATING to 0. Otherwise, just make
 * sure they aren't maxed out. If a cage is sleeping or predictable, also remove said flag and move
 * the cage to the front of active_cages, subtracting its live cell count from sleeping_population
 * or predictable_population[2].
 */
void rattle_cage(cage_type* cage, boolean cage_changed)
{
    if (cage_changed) {
        RESET_TURNS_STABLE(cage);
        RESET_TURNS_OSCILLATING(cage);
    } else {
        if (TURNS_STABLE(cage) == MAX_TURNS_STABLE)
            DEC_TURNS_STABLE(cage);
        else if (TURNS_OSCILLATING(cage) == MAX_TURNS_OSCILLATING)
            DEC_TURNS_OSCILLATING(cage);
    }

    if (IS_ASLEEP(cage)) {
        SET_AWAKE(cage);
        move_cage(cage, &sleeping_cages, &active_cages);
        sleeping_population -= POPULATION_COUNT(cage->bnw[0], cage->bne[0], cage->bsw[0],
                                                cage->bse[0]);
    } else if (IS_PREDICTABLE(cage)) {
        SET_UNPREDICTABLE(cage);
        move_cage(cage,
                  (IS_ONSCREEN(cage) ? &predictable_cages_onscreen : &predictable_cages_offscreen),
                  &active_cages);
        predictable_population[0] -= POPULATION_COUNT(cage->bnw[0], cage->bne[0], cage->bsw[0],
                                                      cage->bse[0]);
        predictable_population[1] -= POPULATION_COUNT(cage->bnw[1], cage->bne[1], cage->bsw[1],
                                                      cage->bse[1]);
    }
}

/* Find and rattle all neighbors of the given cage. Give birth to any neighbors that don't exist
 * yet.
 */
void rattle_neighbors(cage_type* cage)
{
    if (cage->north)
        rattle_cage(cage->north, FALSE);
    else {
        cage->north = find_rattle_birth(cage->x, cage->y-1);
        if (cage->north)
            cage->north->south = cage;
    }
    if (cage->south)
        rattle_cage(cage->south, FALSE);
    else {
        cage->south = find_rattle_birth(cage->x, cage->y+1);
        if (cage->south)
            cage->south->north = cage;
    }
    if (cage->west)
        rattle_cage(cage->west, FALSE);
    else {
        cage->west = find_rattle_birth(cage->x-1, cage->y);
        if (cage->west)
            cage->west->east = cage;
    }
    if (cage->east)
        rattle_cage(cage->east, FALSE);
    else {
        cage->east = find_rattle_birth(cage->x+1, cage->y);
        if (cage->east)
            cage->east->west = cage;
    }
    if (cage->nw)
        rattle_cage(cage->nw, FALSE);
    else {
        cage->nw = find_rattle_birth(cage->x-1, cage->y-1);
        if (cage->nw)
            cage->nw->se = cage;
    }
    if (cage->ne)
        rattle_cage(cage->ne, FALSE);
    else {
        cage->ne = find_rattle_birth(cage->x+1, cage->y-1);
        if (cage->ne)
            cage->ne->sw = cage;
    }
    if (cage->sw)
        rattle_cage(cage->sw, FALSE);
    else {
        cage->sw = find_rattle_birth(cage->x-1, cage->y+1);
        if (cage->sw)
            cage->sw->ne = cage;
    }
    if (cage->se)
        rattle_cage(cage->se, FALSE);
    else {
        cage->se = find_rattle_birth(cage->x+1, cage->y+1);
        if (cage->se)
            cage->se->nw = cage;
    }
}

/* Flag a cage as asleep and move it to sleeping_cages, adding its live cell count to
 * sleeping_population.
 */
void tranquilize_cage(cage_type* cage)
{
    SET_ASLEEP(cage);
    move_cage(cage, &active_cages, &sleeping_cages);
    sleeping_population += POPULATION_COUNT(cage->bnw[0], cage->bne[0], cage->bsw[0],
                                            cage->bse[0]);
}

/* Set or clear the "onscreen" flag for the given cage, depending on viewport settings.
 */
void update_onscreen_flag(cage_type* cage)
{
    int32  x, y;

    x = cage->x * CAGE_SIZE;
    y = cage->y * CAGE_SIZE;
    if (x + CAGE_SIZE > viewport_xstart && x <= viewport_xend &&
        y + CAGE_SIZE > viewport_ystart && y <= viewport_yend) {
        SET_ONSCREEN(cage);
    }
    else
        SET_OFFSCREEN(cage);
}
